"""Implementation of a Chow Classifier"""

from typing import Union
import numpy as np
import pandas as pd
from scipy.stats import f as Fdist
import matplotlib.pyplot as plt
import seaborn as sns
from statsmodels.regression.linear_model import OLS
from statsmodels.tools import add_constant
from .utilities import conf_int_to_trend, parse_cmap
from .utilities import SCATTERPARAMS, FILLPARAMS, FITPARAMS, CI_COLORS, TREND_COLORS

### Set plot style
sns.set_style('white')
plt.rcParams.update({'font.family':'serif',
                     'font.size':14,
                     "xtick.bottom" : True,
                     "ytick.left" : True})
plt.rcParams['font.serif'] = ['EB Garamond', 'Garamond','Times New Roman','Times']


class Chow:
    """
    Class handling the chow analysis
    Call `run()` to run the analysis (find breakpoint and classify)
    Call afterwards `params()` to get the regressions (y=k * x + b) parameters as
    [[model1_k, model1_b, model1_Rsquared],[model2_k, model2_b, model2_Rsquared]]
    (if the breakpoint is not significant, model1 = model2)
    The significance level of the breakpoint is adjusted with Bonferroni correction,
    i.e. dividing alpha by the number of tests. The confidence interval of the slope
    and intercept are at significance level alpha (not corrected).

    :param df: data on which to perform the Chow categorisation
    :param name: name of the dataset (optional)
    :param initial_breakpoint: breakpoint (if None, midpoint of X series will be used)
    :param timecol: name of time column (X)
    :param ycol: name of the variable column (y)
    :param groupcol: name of the group column (g) if not in df will be ignored
    :param alpha: significance level (without Bonferroni correction), 
                  can be changed with ``Chow.set_alpha(0.01)``
    :param margin: range of indexes around breakpoint (or midpoint) where the best breakpoint
                    will be searched. E.g. if X = [0,1,1.5,2,3,4,5], 
                    breakpoint = 2 and margin = 1,
                    the set of breakpoints [1.5,2,3] will be tested.
                    (set = 0 if you want to look only at the given 
                    breakpoint or the midpoint)
    """
    def __init__(self, df,
                name:str = "",
                initial_breakpoint:float = None,
                timecol:str = 'year',
                ycol:str = 'value',
                groupcol:str = 'g',
                alpha = 0.01,
                margin:int = 2):
        """
        Initialize a Chow classifier object.
        
        :param df: DataFrame containing the data to analyze
        :param name: Name identifier for this dataset
        :param initial_breakpoint: Initial breakpoint value 
                                   (if None, uses midpoint)
        :param timecol: Name of the time column in df
        :param ycol: Name of the dependent variable column in df
        :param groupcol: Name of the grouping column in df (optional)
        :param alpha: Significance level for statistical tests
        :param margin: Range around breakpoint to search for optimal break
        """
        try:
            self.df = df.loc[:,[timecol,groupcol, ycol]]
            self.df.columns = ('X','g','y')
        except KeyError:
            self.df = df.loc[:,[timecol,ycol]]
            self.df.columns = ('X','y')
            self.df['g'] = 'a'
        self.name = name
        self.df['X'] = pd.to_numeric(self.df['X'])
        self.df['y'] = pd.to_numeric(self.df['y'])
        self.initial_breakpoint = initial_breakpoint
        self.margin=margin
        self.breakpoints = None
        self.normalise=False
        self.tt = -1
        self.m = margin*2+1 # number of hypotheses tested, will be used for Bonferroni correction
        self.F=None
        self.alpha0 = alpha # original significance level
        self.alpha = None # significance level with Bonferroni correction
        self.model0=None
        self.model1=None
        self.model2=None
        self.gCs = {}# sub analysis with grouping on 'g'
        self.successful=None
        self.color = "#000000"
        self.results = None
        self.ss_res_0 = None
        self.ss_res_1 = None
        self.ss_res_2 = None
        self.score = None
        self.treeclass = None
        self.breakpoint = None
    def get_breakpoint_indices(self):
        """
        Calculate and set the breakpoint indices to search.
        
        :return: List of breakpoint indices to test
        """
        X = sorted(set(self.df['X']))
        if self.breakpoints is not None:
            self.breakpoints = sorted(set(self.breakpoints).intersection(set(X)))
            breakpoint_indices = range(0,len(self.breakpoints))
        else:
            breakpoint_indices = range(0,len(X))
            self.breakpoints = X
        if self.initial_breakpoint is None:
            self.initial_breakpoint = self.breakpoints[int(len(self.breakpoints)/2)]
        if self.initial_breakpoint not in self.breakpoints:
            s = min(range(len(X)), key=lambda i: abs(X[i]-self.initial_breakpoint))
            self.initial_breakpoint = X[s]
        breakpoint_index = self.breakpoints.index(self.initial_breakpoint)
        if breakpoint_index < 3:
            breakpoint_index = 3
            self.initial_breakpoint = self.breakpoints[breakpoint_index]
            print(f"breakpoint index for {self.name} is below 3, will be replace by 3.\n The new breakpoint is {self.initial_breakpoint}")
        elif breakpoint_index > len(X)-4:
            breakpoint_index = len(X)-4
            self.initial_breakpoint = self.breakpoints[breakpoint_index]
            print(f"breakpoint index for {self.name} is above {len(X)-4}, will be replace by {len(X)-4}.\n The new breakpoint is {self.initial_breakpoint}")
        breakpoint_indices = list(range(
            max(3, breakpoint_index - self.margin),
            min(len(X) - 3, breakpoint_index + self.margin) + 1
        ))
        if len(breakpoint_indices) == 0:
            raise ValueError(f"No valid breakpoint indices found for {self.name} with initial breakpoint {self.initial_breakpoint} and margin {self.margin}. Resulted in no breakpoint indices to test.")
        return breakpoint_indices
    def find_best_bkp(self, breakpoint_indices=None, **_):
        """
        Find the best breakpoint and run Chow test to find OLS parameters.
        
        Tests multiple breakpoints and selects the one with the best score.
        Updates self.initial_breakpoint with the optimal value.
        
        :param breakpoint_indices: Optional list of indices to test. 
                                   If None, uses get_breakpoint_indices()
        :param kwargs: Additional keyword arguments (currently unused)
        """
        if breakpoint_indices is not None:
            if not isinstance(breakpoint_indices, list):
                breakpoint_indices=[breakpoint_indices]
            midpoint_index = int(np.median(breakpoint_indices))
        else:
            breakpoint_indices = self.get_breakpoint_indices()
            midpoint_index = breakpoint_indices[int(len(breakpoint_indices)/2)]
        results = {}
        for bk in breakpoint_indices:
            results[bk]=self.run_chow(bkp=self.breakpoints[bk])
        self.results=pd.DataFrame(results)
        self.results = self.results.transpose()
        self.results['delay'] = np.abs(midpoint_index - self.results.index)
        self.results = self.results.sort_values('score')
        #find best breakpoint
        self.initial_breakpoint = self.breakpoints[self.results.score.idxmin()]
        self.m = len(breakpoint_indices)
        self.set_alpha(self.alpha0)
    def run_chow(self, bkp, normalise=False):
        """
        Run Chow test at specified breakpoint.
        
        Fits OLS models to the full dataset and two subsets 
        (before and after breakpoint),
        then calculates the F-statistic to test for structural break.
        
        :param bkp: Breakpoint value at which to split the data
        :param normalise: Whether to normalize the y values before fitting
        :return: Dictionary containing score, F-statistic, 
                 and residual sum of squares
        """
        if normalise:
            self.df['y']=(self.df['y']-self.df['y'].mean())/self.df['y'].std()
        X=self.df.loc[:,'X']
        y=self.df.loc[:,'y']
        X1=self.df.loc[self.df['X']<bkp,'X']
        X2=self.df.loc[self.df['X']>=bkp,'X']
        y1=self.df.loc[self.df['X']<bkp,'y']
        y2=self.df.loc[self.df['X']>=bkp,'y']
        self.successful = False
        if np.any(X1) and np.any(X2):
            X1=add_constant(X1)
            X2=add_constant(X2)
            X=add_constant(X)
            self.model0=OLS(y,X).fit()
            self.model1=OLS(y1,X1).fit()
            self.model2=OLS(y2,X2).fit()
            self.ss_res_0=np.sum((y-self.model0.predict(X))**2)
            self.ss_res_1=np.sum((y1-self.model1.predict(X1))**2)
            self.ss_res_2=np.sum((y2-self.model2.predict(X2))**2)
            k = 2 # Total number of parameters
            self.F=((self.ss_res_0-(self.ss_res_1+self.ss_res_2))/k) / ((self.ss_res_1+self.ss_res_2)/(X.shape[0]-2*k))
            if X.shape[0]-2*k >0:
                self.score=Fdist.sf(self.F, 2, X.shape[0]-2*k)
                self.successful=True
        if self.successful is False and np.any(X) and X.shape[0]>1:
            X=add_constant(X)
            self.model0=OLS(y,X).fit()
            self.model1 = self.model0
            self.model2 = self.model0
            self.ss_res_0=np.sum((y-self.model0.predict(X))**2)
            self.ss_res_1 = None
            self.ss_res_2 = None
            self.F = None
            self.score = 1
            #self.successful = False
        elif self.successful is False:
            raise ValueError(
                f"Data missing for at least one period for {self.name!r},"
                f" with breakpoint {self.initial_breakpoint}"
            )
        return {"score" : self.score,
                "F" : self.F,
                "ss_res_0" : self.ss_res_0,
                "ss_res_1" : self.ss_res_1,
                "ss_res_2" : self.ss_res_2,
                }
    def classify(self, **_):
        """
        Classify the dataset based on results of Chow test at 
        significance level alpha.
        
        Possible classification values:
        
        **No significant breakpoint:**

        1. N: non-significant overall trend
        2. I: significant increasing overall trend
        3. D: significant decreasing overall trend

        **Significant breakpoint** (set1 and set2 indicate points 
        before/after breakpoint):

        4. NN: non-significant trend on set1 and non-significant trend on set2
        5. NI: non-significant trend on set1 and significant increasing
               trend on set2
        6. ND: non-significant trend on set1 and significant decreasing
               trend on set2
        7. IN: significant increasing trend on set1 and non-significant
               trend on set2
        8. ID: significant increasing trend on set1 and significant
               decreasing trend on set2
        9. iI: significant increasing trend on both set1 and set2
               with greater increase in set2
        10. Ii: significant increasing trend on both set1 and set2
                with greater increase in set1
        11. DN: significant decreasing trend on set1 and non-significant
                trend on set2
        12. DI: significant decreasing trend on set1 and significant
                increasing trend on set2
        13. dD: significant decreasing trend on both set1 and set2 with
                greater decrease in set2
        14. Dd: significant decreasing trend on both set1 and set2 with
                greater decrease in set1
        
        :param kwargs: Additional keyword arguments (currently unused)
        :return: Dictionary with model parameters
        """
        if self.alpha is None:
            self.set_alpha(self.alpha0)
        tc = 'N'
        if self.score < self.alpha:#
            tt1= conf_int_to_trend(self.model1.conf_int(alpha = self.alpha0))
            tt2= conf_int_to_trend(self.model2.conf_int(alpha = self.alpha0))
            # correction for insignificant trend on one period when actually
            # the trend is significantly higher or lower than the next/previous period
            # at the initial_breakpoint year (i.e. confidence intervals are disjoints at that point)
            if (tt1==0 or tt2==0) and self.initial_breakpoint:
                lci1, uci1 = self.model1.conf_int(alpha = self.alpha0).loc['X',:]
                lci2, uci2 = self.model2.conf_int(alpha = self.alpha0).loc['X',:]
                b1, a1 = self.model1.params
                b2, a2 = self.model2.params
                right_limit = b1 + self.initial_breakpoint * a1  #right limit of first period
                left_limit = b2 + self.initial_breakpoint * a2#left limit of second period
                if right_limit >= left_limit:
                    if b2 + lci2 * self.initial_breakpoint > b1 + uci1 * self.initial_breakpoint:
                        if tt1==0:
                            tt1=1 # the trend is in fact increasing
                        if tt2==0:
                            tt2=2 # the trend is in fact decreasing
                else:
                    if b2 + uci2 * self.initial_breakpoint < b1 + lci1 * self.initial_breakpoint:
                        if tt1==0:
                            tt1=2 # the trend is in fact decreasing
                        if tt2==0:
                            tt2=1 # the trend is in fact increasing
            if tt1==0:
                tc=['NN','NI','ND'][tt2]
            elif tt1==1:
                if tt2==1:
                    tc = ['iI','Ii'][1*(self.model1.params['X']>self.model2.params['X'])]
                else:
                    tc=['IN','','ID'][tt2]
            elif tt1==2:
                if tt2==2:
                    tc = ['dD','Dd'][1*(self.model1.params['X']<self.model2.params['X'])]
                else:
                    tc = 'DN' if tt2 == 0 else 'DI'
        else:
            #N0, N1 or N2
            self.tt= conf_int_to_trend(self.model0.conf_int(alpha = self.alpha0))
            tc = ['N','I','D'][self.tt]
            self.initial_breakpoint=None
        self.treeclass = tc
        self.set_color()
        return self.params()
    def set_color(self):
        """
        Set the color attribute based on the trend type (tt).
        
        Uses predefined color scheme from TREND_COLORS.
        """
        self.color = TREND_COLORS[self.tt]
    def params_names(self):
        """
        Get list of parameter names for model output.
        
        :return: List of parameter name strings
        """
        return ['intercept','slope',
                'R2','se_intercept','se_slope','n',
                'intercept_pvalue', 'slope_pvalue',
                'intercept_ci_lower','intercept_ci_upper',
                'slope_ci_lower','slope_ci_upper',
                'alpha','alpha_Bonferroni_corrected']
    def params(self):
        """
        Return parameters of the fitted model(s).
        
        :return: Dictionary containing model0, model1, and model2 parameters.
                 If breakpoint is significant, model0 is empty and model1/model2 are populated.
                 Otherwise, model0 is populated and model1/model2 are empty.
        """
        def get_params(model):
            d = [*list(model.params),\
                model.rsquared,\
                *list(model.bse),\
                self.df.shape[0],\
                *list(model.pvalues),\
                *list(model.conf_int(alpha = self.alpha0).loc['const',:]),\
                *list(model.conf_int(alpha = self.alpha0).loc['X',:]),\
                self.alpha0,\
                self.alpha]
            return {k:v for v,k in zip(d,self.params_names())}
        if self.score < self.alpha:
            return {'model0':{},
                    'model1':get_params(self.model1),
                    'model2':get_params(self.model2)}
        return {'model0':get_params(self.model0),
                'model1':{},
                'model2':{}}
    def run(self, **kwargs):
        """
        Run the complete Chow classification analysis.
        
        Finds best breakpoint, runs Chow test, and classifies the trend.
        
        :param kwargs: Keyword arguments forwarded to find_best_bkp and classify.
                      Can include 'normalise' for run_chow, 'alpha' for classify.
        :return: Classification code (string) or None if analysis fails
        """
        if np.any(self.df):
            try:
                self.find_best_bkp(**kwargs)
            except IndexError:
                print(f"No suitable breakpoint found for {self.name}.")
            self.run_chow(bkp = self.initial_breakpoint, normalise = kwargs.get('normalise',False))
            try:
                self.breakpoint = self.initial_breakpoint if self.score < self.alpha else None
            except Exception as e:
                raise ValueError(f"Error determining breakpoint for {self.name}, {self.score=}, {self.alpha=}: {e}")
            self.classify(**kwargs)
            return self.treeclass
        else:
            self.successful=False
            return None
    def set_alpha(self, alpha:float):
        """
        Change the current statistical significance level.
        
        The Bonferroni correction will be applied based on the number of tests.

        :param alpha: Significance level, positive float in (0,1)
        """
        self.alpha0 = alpha
        self.alpha = self.alpha0 / max(self.m,1) # applying bonferroni correction

    def summary(self):
        """
        Print the statistical summary of the fitted model(s).
        
        If breakpoint is significant, prints summaries for both model1 and model2.
        Otherwise, prints summary for model0 (whole dataset).
        """
        if self.score < self.alpha:
            print(f"Summary model 1: start-{self.initial_breakpoint}")
            self.model1.summary()
            print(f"Summary model 2: {self.initial_breakpoint}-end")
            self.model2.summary()
        else:
            print("Summary model 0: whole dataset")
            self.model0.summary()
    def plot(self, filename:str = None, ax = None, figsize = (16,8),
            ylog:bool=False,
            fill:bool = True,
            scatter:bool = True,
            scatterparams:dict = None,
            fillparams:dict = None,
            fitparams:dict = None,
            linestyles:list = None,
            show_legend:bool = True,
            **params):
        """
        Plot the regression model(s).

        :param filename: Filename (with extension format), if None, will call plt.show() instead
        :param ax: Matplotlib axis on which to plot, if None will create new figure
        :param figsize: Tuple with figure size
        :param ylog: Make the y-axis log-scale
        :param fill: Plot the confidence interval
        :param scatter: Plot individual points
        :param scatterparams: Parameters for the scatter plot
        :param fillparams: Parameters for the confidence intervals
        :param fitparams: Parameters for the fitted line
        :param linestyles: Select the linestyle of each trend (list of size 3), set to None if you want to define it in fitparams
        :param show_legend: Legend options passed to ax.legend()
        :param params: Additional parameters including color, xlabel, ylabel, title, xlim, ylim
        """
        ### update plot parameters
        SCATTERPARAMS.update(scatterparams if scatterparams is not None else {})
        FILLPARAMS.update(fillparams if fillparams is not None else {})
        FITPARAMS.update(fitparams if fitparams is not None else {})
        # record if it is plotting for parent process
        main_p = ax is None
        # create the plot:
        self.set_color()
        def add_reg_plot(sub_df, model):
            # X values
            X = sorted(set(sub_df['X']))
            inc = (max(X)-min(X))/max(1,len(X)-1)
            X = np.linspace(min(X)-0.5*inc, max(X)+0.5*inc, 2*(len(X)+1))
            Xc = add_constant(X)
            # get confidence interval
            ci_l,ci_u = zip(*model.get_prediction(Xc).conf_int(alpha = self.alpha))
            tt = conf_int_to_trend(model.conf_int(alpha = self.alpha))
            if fill is True:
                # set color depending on time trend
                FILLPARAMS['color'] = fillparams.get('color',CI_COLORS[tt])   
                ax.fill_between(x = X, y1=ci_l, y2 = ci_u, **FILLPARAMS)
            FITPARAMS['color'] = fitparams.get('color',TREND_COLORS[tt])
            if linestyles is not None:
                if isinstance(linestyles, list) and len(linestyles)==3:
                    FITPARAMS['linestyle'] = linestyles[tt]
                elif isinstance(linestyles, str):
                    FITPARAMS['linestyle'] = linestyles
                else:
                    raise ValueError("linestyles should be either a string or a list of 3 linestyles for each trend category (neutral, increasing, decreasing).")
            else:
                FITPARAMS['linestyle'] = FITPARAMS.get('linestyle','solid')
            intercept, slope = model.params
            y0 = intercept + slope * min(X)
            y1 = intercept + slope * max(X)
            ax.plot((min(X),max(X)),(y0,y1), marker = None, **FITPARAMS)
        if ax is None:
            plt.clf()
            f, (ax) = plt.subplots(nrows=1, ncols=1, figsize=figsize)
        if self.score < self.alpha:
            df1 = self.df.loc[self.df['X'] <  self.initial_breakpoint,]
            df2 = self.df.loc[self.df['X'] >= self.initial_breakpoint,]
            add_reg_plot(df1,self.model1)
            add_reg_plot(df2,self.model2)
        else:
            add_reg_plot(self.df,self.model0)
        if scatter is True:
            SCATTERPARAMS['color'] = scatterparams.get('color',self.color)
            ax.scatter(self.df['X'],self.df['y'],
                        **SCATTERPARAMS)
        # Tweak the visual presentation
        ax.xaxis.grid(True)
        ax.yaxis.grid(True)
        ax.minorticks_on()
        ax.set(ylabel=params.get("ylabel", 'y'),
            xlabel=params.get("xlabel",'X'),
            title=params.get("title",self.name))
        ax.set_xlim(*params.get('xlim',(self.df['X'].min(), self.df['X'].max())))
        if params.get('ylim') is not None:
            ax.set_ylim(*params.get('ylim',(self.df['y'].min(), self.df['y'].max())))
        if ylog is True:
            ax.set_yscale('log')
        if show_legend is False:
            ax.get_legend().remove()
        #sns.despine(trim=True, left=True)
        #f.suptitle("")
        if filename is not None and main_p is True:
            plt.savefig(f'{filename}')
            plt.close()
        elif main_p is True:
            plt.show()
            plt.close()
    def plot_by_group(self,
                      filename:str = None,
                      ax = None,
                      figsize = (16,8),
                      cmap:str = 'Set1',
                      show_legend:bool=True,
                      plot_overall:bool=True,
                      plot_individual_fill:bool = True,
                      scatterparams:dict = None,
                      fillparams:dict = None,
                      fitparams:dict = None,
                      groups_order:list = None,
                      **params):
        """
        Plot the regression model(s) grouped by the group column.

        :param filename: Filename (with extension format), if None, will call plt.show() instead
        :param ax: Matplotlib axis on which to plot, if None will create new figure
        :param figsize: Tuple with figure size
        :param cmap: Colormap name (from matplotlib) or custom color mapping
        :param show_legend: Include the legend
        :param plot_overall: Plot overall trend and confidence interval
        :param plot_individual_fill: Plot confidence interval for each individual group
        :param scatterparams: Parameters for the scatter plot
        :param fillparams: Parameters for the confidence intervals
        :param fitparams: Parameters for the fitted line
        :param groups_order: List showing the order in which the groups must be plotted
        :param params: Additional parameters including xlabel, ylabel, title, ylog, scatter
        """
        cmap = parse_cmap(cmap)
        scatterparams = scatterparams or {}
        fillparams = fillparams or {}
        fitparams = fitparams or {}
        params['xlabel'] = params.get('xlabel','X')
        params['ylabel'] = params.get('ylabel','y')
        # record if it is plotting for parent process
        main_p = ax is None
        if ax is None:
            plt.clf()
            _, (ax) = plt.subplots(nrows=1,
                    ncols=1,
                    sharey=params.get("sharey",True),
                    figsize=params.get('figsize',figsize))
        if groups_order is None:
            groups_order = []
        params['title'] = self.name
        groups_order = groups_order + [x for x in set(self.df['g']) if x not in groups_order]
        self.df['go'] = [groups_order.index(x) for x in self.df['g']]
        for i,(n,dfg) in enumerate(self.df.groupby(['go','g'],sort=True)):
            n = n[1]
            if n not in self.gCs:
                self.gCs[n] = Chow(df = dfg,
                            name = n,
                            initial_breakpoint = self.initial_breakpoint,
                            timecol = 'X',
                            ycol = 'y',
                            groupcol = 'g',
                            alpha = self.alpha,
                            margin = self.margin)
            success = self.gCs[n].run()
            if success is None:
                print(f"Analysis failed for {self.name} {n}")
            else:
                scatterparams['color'] = scatterparams.get('color', cmap.get(n,cmap[i]))
                fitparams['color'] = fitparams.get('color', cmap.get(n,cmap[i]))
                fillparams['color'] = fillparams.get('color', cmap.get(n,cmap[i]))
                scatterparams['label'] = n
                self.gCs[n].plot(ax = ax,
                    fill = plot_individual_fill,
                    scatterparams = scatterparams,
                    fillparams = fillparams,
                    fitparams = fitparams,
                    **params)
        if plot_overall is True:
            scatterparams['color'] = (0,0,0)
            fitparams['color'] = (0,0,0)
            fillparams['color'] = (0,0,0)
            scatterparams['label'] = params.get('overall_label','Overall')
            self.plot(ax = ax,
                        scatter = False,
                        scatterparams = scatterparams,
                        fillparams = fillparams,
                        fitparams = fitparams,
                        **params)
        if show_legend is True:
            ax.legend()
        if filename is not None and main_p is True:
            plt.savefig(f'{filename}')
            plt.close()
        elif main_p is True:
            plt.show()
            plt.close()


import numpy as np
import pandas as pd
from chowclassifier import Chow
import sys
import os
sys.path.append(os.path.abspath('../src'))

class small_noise_gen():
    def __init__(self, mu=0.,sigma=1e-4, n = 10000):
        self.mu = mu
        self.sigma = sigma
        self.size = n
        self.set_noise()
    def set_noise(self):
        self.noise = list(np.random.normal(loc=self.mu,scale=self.sigma, size=self.size))
    def __next__(self):
        try:
            return self.noise.pop()
        except IndexError:
            self.set_noise()
            return self.noise.pop()


def test_chow():
    """Expects perfect categorisation on quasi-perfectly aligned data"""
    np.random.seed(42)
    X = range(0,100)
    cl = {'iI':{'bkp':50,'A':[1,2]},
            'Ii':{'bkp':50,'A':[2,1]},
            'dD':{'bkp':50,'A':[-1,-2]},
            'Dd':{'bkp':50,'A':[-2,-1]},
            'ID':{'bkp':50,'A':[1,-1]},
            'DI':{'bkp':50,'A':[-1,1]},
            'IN':{'bkp':50,'A':[1,0]},
            'NI':{'bkp':50,'A':[0,1]},
            'DN':{'bkp':50,'A':[-1,0]},
            'ND':{'bkp':50,'A':[0,-1]},
            'NN':{'bkp':50,'A':[0,0]},
            'N':{'bkp':None,'A':0},
            'D':{'bkp':None,'A':-1},
            'I':{'bkp':None,'A':1}}
    sng = small_noise_gen()
    p = 0
    i = 0
    for k,v in cl.items():
        if v['bkp'] is None:
            df = pd.DataFrame(zip(X,[v['A']*x*2+0 + next(sng) for x in X]), columns = ['year','value'])
        else:
            if k == 'NN':
                b = 1
            else:
                b = v['bkp'] * v['A'][0]
            y = [v['A'][0]*x*2 + 0 + next(sng) for x in X if x < v['bkp']] +\
                      [v['A'][1]*x*2 + b + next(sng) for x in X if x >= v['bkp']]
            df = pd.DataFrame(zip(X,y), columns = ['year','value'])
        C = Chow(df = df,margin = 50)
        C.run()
        p+=int(C.treeclass == k)
        if C.treeclass != k:
            print(f"Expected {k}, got {C.treeclass}")
        i+=1
    assert p/float(i) > 0.99, f"Divergence in test_chow, found true class with rate {p}."
    
    

            


"""Command-line interface for the chowclassifier package."""
# __main__.py
import argparse
from chowclassifier.chewdata import ChewData


def main():
    """
    run analysis on given data

    :params -f, --filename: filename
    :params -s, --sep: separator for parsing the filename (if type csv)
    :params -X, --timecol: timecol
    :params -y, --ycol: ycol
    :params -n, --namecol: name of column for groupby
    :params -b, --breakpoint: initial breakpoint (if None, midpoint will be taken)
    :params -k, --breakpoints: candidates for breakpoints (list)
    :params -m, --margin: margin around initial breakpoint
    :params -o, --output: filename for the excel or csv file output
    :params -p, --path: path where the figures will be saved
    :params -a, --alpha: significance level in (0,1)
    :params --no-figures: Do not export figures
    :params --no-table: Do not export results table
    :params --nrows: Numbers of rows in the grouped figure (leave ncols blank)
    :params --ncols: Number of columns in the groupe figure (leave nrows blank)
    :params --fig-title: title of the figure
    :params --xlabel: label of the x axis
    :params --ylabel: label of the y axis
    :params --sharey: share y axis across subplots
    :params --figsize: size of figures
    """
    parser = argparse.ArgumentParser()
    parser.add_argument('-f','--filename', type=str)
    parser.add_argument('-X','--timecol', type=str, default = 'date')
    parser.add_argument('-y','--ycol', type=str, default = 'value')
    parser.add_argument('-n','--namecol', type=str, default = 'name')
    parser.add_argument('-b','--breakpoint', type=float, default = None)
    parser.add_argument('-k','--breakpoints', type=list, default = None)
    parser.add_argument('-m','--margin', type=int, default = 5)
    parser.add_argument('-o','--output', type=str, default = 'output.xlsx')
    parser.add_argument('-p','--path', type=str, default = 'figs')
    parser.add_argument('-s','--sep', type=str, default = ',')
    parser.add_argument('-a','--alpha', type=float, default = 0.01)
    parser.add_argument('--no-figures', type=bool, default = False)
    parser.add_argument('--no-table', type=bool, default = False)
    parser.add_argument('--nrows', type=int, default = None)
    parser.add_argument('--ncols', type=int, default = 3)
    parser.add_argument('--fig-title', type=str, default = '')
    parser.add_argument('--fig-format', type=str, default = 'pdf')
    parser.add_argument('--xlabel', type=str, default = 'X')
    parser.add_argument('--ylabel', type=str, default = 'y')
    parser.add_argument('--sharey', type=bool, default = True)
    parser.add_argument('--figsize',type = list, default = (16,8))
    args = parser.parse_args()
    if args.filename is None:
        print("No data given. Please pass filename as argument --filename 'data.csv'")
        return
    chew = ChewData(filename = args.filename,
                namecol = args.namecol,
                ycol = args.ycol,
                timecol = args.timecol,
                sep = args.sep,
                margin = args.margin,
                alpha = args.alpha,
                initial_breakpoint = args.breakpoint)
    chew.run()
    if args.no_table is False:
        chew.save_to_file(filename = args.output, sep = args.sep)
    if args.no_figures is False:
        chew.plot(f'output.{args.fig_format}',
               nrows = args.nrows,
               ncols = args.ncols,
               sharey = args.sharey,
               figsize = args.figsize
               )
        chew.plot_individually(savingpath = args.path, format = args.fig_format)


if __name__ == "__main__":

    main()



"""Chow classifier package for time-series trend classification."""

# __init__.py

# Version of the chowclassifier package
__version__ = "1.0.9"
from chowclassifier.chow import Chow
from chowclassifier.chewdata import ChewData



import os
def test_all_examples():
    REPO_ROOT = os.path.dirname(os.path.dirname(__file__))
    example_dir = os.path.join(REPO_ROOT, 'example')
    os.system(f"python {os.path.join(example_dir, 'example.py')}")


"""Tests"""
import string
import numpy as np
import pandas as pd
from chowclassifier import ChewData
import sys
import os
sys.path.append(os.path.abspath('../src'))

def labels(alphabet=string.ascii_uppercase):
    assert len(alphabet) == len(set(alphabet))  # make sure every letter is unique
    s = [alphabet[0]]
    while 1:
        yield ''.join(s)
        l = len(s)
        for i in range(l-1, -1, -1):
            if s[i] != alphabet[-1]:
                s[i] = alphabet[alphabet.index(s[i])+1]
                s[i+1:] = [alphabet[0]] * (l-i-1)
                break
        else:
            s = [alphabet[0]] * (l+1)

class yield_param():
    def __init__(self,func,**params):
        self.A = list(func(**params)) # generate param
    def __next__(self):
        return self.A.pop(0)

def synt_series(X,
                a, b, r,
                seed = None):
    """Generate a random series"""
    if seed:
        np.random.seed(seed)
    mu = 0
    sigma = np.random.gamma(1,1,1)
    ng = np.random.normal(mu, sigma, len(X)*r)
    return [x*a+b for x in X for i in range(0,r)] + ng

def classify(a,b,bkp, z = 1e-2):
    if bkp is not None:
        if a[0]>z and a[1]>z and a[0]>a[1]:
            return 'Ii'
        elif a[0]>z and a[1]>z and a[0]<=a[1]:
            return 'iI'
        elif a[0]<-z and a[1]<-z and a[0]>a[1]:
            return 'Dd'
        elif a[0]<-z and a[1]<-z and a[0]<=a[1]:
            return 'dD'
        elif a[0]>z and a[1]<-z:
            return 'ID'
        elif a[0]>z and a[1]>-z and a[1]<z:
            return 'IN'
        elif a[0]<-z and a[1]>-z and a[1]<z:
            return 'DN'
        elif a[0]<-z and a[1]>z:
            return 'DI'
        elif a[1]>z and a[0]>-z and a[0]<z:
            return 'NI'
        elif a[1]<-z and a[0]>-z and a[0]<z:
            return 'ND'
        elif a[0]>-z and a[0]<z and a[1]>-z and a[1]<z:
            return 'NN'
    else:
        if a > z:
            return 'I'
        elif a < -z:
            return 'D'
        else:
            return 'N'

def synt_data(n=1000,m=50,seed = None):
    X = np.linspace(0,n,2*n+2)
    labs = labels()
    r = 3
    data = {'year':[x for x in X for i in range(0,r)]}
    exp = {}
    with_breakpoints = np.random.binomial(1,0.7,m)# flip coin to see if broken at bkp (favour breaking)
    bks = [x if y == 1 else None for x,y in zip(np.random.choice(range(5,len(X)-5),m),with_breakpoints)]# select random breakpoints
    nparams = m + sum(with_breakpoints) # total number of pairs of parameters needed
    A = yield_param(np.random.normal,loc=0,scale=2,size=nparams) # slope gen
    b = yield_param(np.random.poisson,lam=15,size=m) # intercept gen
    for i in range(0,m):
        l = next(labs)
        bk = bks[i]
        if bk is not None:
            a1, b1 = next(A),next(b)
            a2 = next(A)
            if np.abs(a1-a2) < 0.15: # make sure the two slopes are different enough to be detected
                a2 += 0.15 if a1>=0 else -0.15
            b2 = a2 * bk + b1
            s1 = synt_series(X[:bk],a1,b1,r,seed = seed)
            s2 = synt_series(X[bk:],a2,b2,r,seed = seed)
            series = np.concatenate((s1,s2))
            tt = classify((a1,a2),(b1,b2),bkp=bk)
            bk = X[bk]
        else:# else no breakpoint
            bk = None
            # generate random slope and intercept
            a0 = next(A)
            b0 = next(b)
            tt = classify(a0,b0,bkp=bk)
            # generate random series with picked slope and intercept
            series =synt_series(X,a0,b0,r,seed = seed)
        data[l] = series 
        exp[l] = {'tt':tt,'bkp':bk}
    df = pd.DataFrame(data)
    return df,exp

def test_synt_data():
    np.random.seed(42)
    sdf,exp = synt_data(n=40,m=50)
    C = ChewData(sdf, margin = 40)
    C.run(initial_breakpoint=None)
    res = pd.DataFrame(columns = ['tt','bk','ett','ebk','acc'])
    for c in C.chows:
        if c.treeclass == exp[c.name]['tt']:
            acc = 0
        elif c.breakpoint is not None and exp[c.name]['bkp'] is None:
            if c.treeclass.upper()[0]==exp[c.name]['tt']:
                if c.treeclass.upper()[1]==exp[c.name]['tt']:
                    acc = 1
                else:
                    acc = 2
            elif c.treeclass.upper()[1]==exp[c.name]['tt']:
                acc = 2
            else:
                acc = 4
        elif c.breakpoint is None and exp[c.name]['bkp'] is not None:
            if exp[c.name]['tt'].upper()[0]==c.treeclass:
                if exp[c.name]['tt'].upper()[1]==c.treeclass:
                    acc = 1
                else:
                    acc = 2
            elif exp[c.name]['tt'].upper()[1]==c.treeclass:
                acc = 2
            else:
                acc = 4
        elif c.breakpoint is not None and exp[c.name]['bkp'] is not None:
            if exp[c.name]['tt'].upper()[0]==c.treeclass.upper()[0]:
                if exp[c.name]['tt'].upper()[1]==c.treeclass.upper()[1]:
                    acc = 1
                else:
                    acc = 2
            elif exp[c.name]['tt'].upper()[1]==c.treeclass.upper()[1]:
                acc = 2
            else:
                acc = 4
        else:
            acc = 4
        res.loc[len(res)] = [c.treeclass, c.breakpoint,
                                exp[c.name]['tt'],#expected class
                                exp[c.name]['bkp'],#expected breakpoint
                                acc] # acc = 0 means correct assignment.
    x_step = 40 / 81  # linspace step size for n=40 (82 points over range 0-40)
    res['acc_bk'] = [abs(v['bk'] - v['ebk']) <= 5 * x_step if v['bk'] is not None and v['ebk'] is not None else True for i,v in res.iterrows()]
    breakpoint_acc = sum(res['acc_bk'])/len(res)
    print(f"Breakpoint accuracy: {breakpoint_acc:.2%}")
    assert breakpoint_acc > 0.70, f"Got wrong breakpoint accurary, {sum(res['acc_bk'])/len(res)}"
    assert sum(res['acc'])/len(res) < 0.6, f"Got many missmatch {res.loc[res['tt']!=res['ett'],:]}"





"""Implementation of a Chow Classifier"""

import os
from typing import Union
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from tqdm import tqdm
from .utilities import infer_date_format, guess_timelapse_type
from .chow import Chow


class ChewData():
    """
    Handle a full data frame with grouping of time series.
    
    :param df: Dataset as a pandas DataFrame (optional if filename provided)
    :param filename: Path to dataframe file (used if df not given)
    :param namecol: Name of the column with group names (used in groupby)
    :param initial_breakpoint: Initial breakpoint value
    :param sep: Separator character in CSV files
    :param timecol: Name of the time column in dataset
    :param ycol: Name of the variable column in dataset
    :param groupcol: Name of the grouping column in dataset
    :param margin: Range of indexes around breakpoint/midpoint, 
                   see Chow.get_breakpoint_indices
    :param alpha: Significance level at which to test the trends
                  (Bonferroni correction will be applied later)
    """
    def __init__(self, 
                df:pd.DataFrame = None, 
                filename:str = None,
                groupcol:Union[None,str] = 'g',
                namecol:str = 'name',
                initial_breakpoint:float = None,
                sep=',',
                timecol:str = 'year',
                ycol:str = 'value',
                margin:int = 2,
                alpha:float = 0.01):
        """
        Initialize a ChewData object for analyzing multiple time series.
        
        Either df or filename must be provided. The class will automatically
        detect wide vs long format and convert if necessary.
        """
        # get dataframe
        
        self.df = df
        self.namecol = namecol
        self.timecol = timecol
        self.groupcol = groupcol
        if ycol is None:
            self.ycol = 'value'
        else:
            self.ycol = ycol
        self.margin = margin
        self.alpha = alpha
        self.initial_breakpoint = initial_breakpoint
        self.chows = []
        self.failed = []
        self.results = None
        if self.df is None:
            self.get_df(filename, sep)
        if self.namecol not in self.df.columns and\
            len([k for k in self.df.columns if k not in ['id',ycol,timecol,'',np.nan]])>0:
            #inferred that df is in wide format
            try:
                self.melt()
            except ValueError as e:
                raise ValueError(
                    "Could not melt the dataframe. Is it really in wide format?"
                    " Check given namecol"
                ) from e
        if pd.api.types.is_numeric_dtype(self.df[self.timecol]) is False:
            date_format = infer_date_format(self.df.loc[self.df[self.timecol]!=np.nan,self.timecol][0])
            print(f'Inferred format {date_format} for column {self.timecol}')
            if date_format is not None:
                self.parse_timecol(date_format)
    def melt(self):
        """
        Pivot the dataframe from wide to long format.
        
        Converts a wide-format dataframe (with time series in columns) to
        long format (with a variable column for series names).
        """
        if self.groupcol in self.df.columns:
            self.df = self.df.melt(id_vars=[self.timecol,self.groupcol], value_name=self.ycol).dropna()
        else:
            self.df = self.df.melt(id_vars=[self.timecol], value_name=self.ycol).dropna()
        self.namecol = 'variable'
    def get_df(self, filename, sep):
        """
        Load dataframe from file.
        
        Supports Excel (.xlsx, .xls), CSV (.csv), and TSV (.tsv) formats.
        
        :param filename: Path to the file
        :param sep: Separator character for CSV files
        :raises ValueError: If file extension is not recognized
        """
        ext = filename.split('.')[-1]
        if ext in ('xlsx', 'xls'):
            self.df = pd.read_excel(filename, engine='openpyxl')
        elif ext == 'csv':
            self.df = pd.read_csv(filename,sep = sep)
        elif ext == 'tsv':
            self.df = pd.read_csv(filename, sep='\t')
        else:
            raise ValueError(f"Filename {filename} does not have a recognised extension")
    def parse_timecol(self, date_format:str = '%Y-%m-%d'):
        """
        Transform time column from dates to numeric values.
        
        The time column must be numeric for analysis. If you have dates instead,
        use this method to transform that column. Note: you need to adjust the
        breakpoint accordingly!

        :param date_format: Date format string (default '%Y-%m-%d')
        """
        series,self.datetype = guess_timelapse_type(self.df[self.timecol], date_format = date_format)
        self.df[self.datetype] = series
        self.timecol = self.datetype
        print(f'Indexed the time column {self.timecol} as {self.datetype}s')
    def run_single(self,
                   sub_df:pd.DataFrame,
                   name:str,
                   initial_breakpoint:Union[float,int])->(dict,Chow):
        """
        Run the Chow classification on a single dataframe.
        
        :param sub_df: DataFrame containing the time series data
        :param name: Name identifier for the series
        :param initial_breakpoint: Initial break point for Chow test 
                                   (if None, will take midpoint of timecol)
        :return: Tuple of (dictionary of parameters from the analysis,
                 Chow object from the analysis)
        """
        chow = Chow(df = sub_df,
                name = name,
                initial_breakpoint = initial_breakpoint,
                timecol = self.timecol,
                ycol = self.ycol,
                alpha = self.alpha,
                margin = self.margin)
        try:
            chow.run()
        except ValueError:
            return None, None
        params = chow.params() # get the regressions parameters
        params['name'] = name
        return params, chow
    def run(self, initial_breakpoint = None):
        """
        Run the Chow classification for each group in the dataset.
        
        :param initial_breakpoint: Initial break point for Chow test
                                  (if None, will take midpoint of timecol)
        """
        results = []
        self.chows = []
        self.failed = []
        if initial_breakpoint is None and self.initial_breakpoint is None:
            x = sorted(set(self.df[self.timecol]))
            self.initial_breakpoint = x[int(len(x)/2)]
            initial_breakpoint = self.initial_breakpoint
            print(f"Set {initial_breakpoint = }")
        elif initial_breakpoint is None:
            initial_breakpoint = self.initial_breakpoint
        print("Running analysis")
        with tqdm(total = len(self.df.groupby(self.namecol))) as pbar:
            for name, sub_df in self.df.groupby(self.namecol):
                if sub_df.shape[0] > 1:
                    params, chow = self.run_single(
                        sub_df, name=name, initial_breakpoint=initial_breakpoint
                    )
                    if chow is None:
                        self.failed.append(name)
                    else:
                        results.append(params)
                        self.chows.append(chow)
                else:
                    self.failed.append(name)
                pbar.update(1)
        self.results = pd.DataFrame(results) # put the results in a dataframe
    def plot(self, filename:Union[str,None]=None,
             grouped:bool = False,
             divide:int = 1,
             nrows:Union[None,int] = None,
             ncols:Union[None,int] = None,
             cmap:Union[None,str] = None,
             scatterparams:dict = None,
             fillparams:dict = None,
             fitparams:dict = None,
             **params):
        """
        Plot all trends in one image.
        
        :param filename: Name of the output file with extension
                         (e.g., 'name.pdf')
        :param grouped: Whether to plot overall trends or grouped trends
        :param divide: Divide the plots into n files
        :param nrows: Number of rows in subplot grid
        :param ncols: Number of columns in subplot grid
        :param cmap: (grouped) Name of one of matplotlib cmaps
        :param scatterparams: Dictionary of parameters for the scatter 
                              part of the plot
        :param fillparams: Dictionary of parameters for the error fill 
                           area of the plot
        :param fitparams: Dictionary of parameters for the fitted line
        :param params: Named parameters passed to matplotlib 
                       (e.g., xlabel, ylabel, sharey, figsize, pad,
                       plot_overall, plot_individual_fill, title)
        """
        scatterparams = scatterparams or {}
        fillparams = fillparams or {}
        fitparams = fitparams or {}
        def get_n(m, _nplots=len(self.chows)):
            n = _nplots / m
            if n - int(n) == 0.:
                return int(n)
            return int(n) + 1
        if filename is not None:
            f = filename.split(".")
            filename = ".".join(f[0:-1])
            ext = f[-1]
        nplots = get_n(divide)
        print("Plotting trends")
        with tqdm(total = divide) as pbar:
            for i in range(0, divide):
                if nrows is None and ncols is None:
                    ncols = 4
                    nrows = get_n(ncols,nplots)
                elif nrows is not None:
                    ncols = get_n(nrows,nplots)
                else:
                    nrows = get_n(ncols,nplots)
                f, axes = plt.subplots(nrows=nrows,
                                    ncols=ncols,
                                    sharey=params.get("sharey",True),
                                    figsize=params.get('figsize',(16,8)))
                f.tight_layout(pad=params.get('pad',3.0)) # increase padding between sub figures
                for chow_obj, ax in zip(self.chows[i*nplots:(i+1)*nplots]+[None for _ in range(ncols*nrows - len(self.chows[i*nplots:(i+1)*nplots]))], axes.reshape(-1)):
                    if chow_obj is None:
                        ax.axis('off')
                        continue
                    if grouped is True and self.groupcol in chow_obj.df.colnames:
                        plt.clf()
                        chow_obj.plot_by_group(ax = ax,
                                              xlabel=params.get('xlabel',self.timecol),
                                              ylabel=params.get('ylabel',self.ycol),
                                              cmap = cmap,
                                              title=chow_obj.name,
                                              scatterparams = scatterparams,
                                              fillparams = fillparams,
                                              fitparams = fitparams,
                                              **params)
                    else:
                        chow_obj.plot(ax=ax,
                            xlabel=params.get('xlabel','X'),
                            ylabel=params.get('ylabel','y'),
                            title=chow_obj.name,
                            scatterparams = scatterparams,
                            fillparams = fillparams,
                            fitparams = fitparams)
                if filename is not None:
                    plt.savefig(f'{filename}_{i:02d}.{ext}')
                pbar.update(1)
        if filename is None:
            plt.show()
            plt.close()
    def plot_individually(self, savingpath:str = 'figs', fmt:str = 'pdf', **params):
        """
        Plot all trends in individual files.
        
        :param savingpath: Path where the plots will be saved
        :param fmt: Format of the image (e.g., 'pdf', 'png')
        :param params: Named parameters passed to matplotlib
                       (e.g., xlabel, ylabel, prefix, scatterparams,
                       fillparams, fitparams)
        """
        # check if saving path exists:
        if not os.path.isdir(savingpath):
            os.makedirs(savingpath)
        print("Plotting trends individually")
        with tqdm(total = len(self.chows)) as pbar:
            params['xlabel'] = params.get('xlabel', self.timecol)
            params['ylabel'] = params.get('ylabel', self.ycol)
            params['scatterparams'] = params.get('scatterparams', {})
            params['fillparams'] = params.get('fillparams', {})
            params['fitparams'] = params.get('fitparams', {})
            for chow_obj in self.chows:
                plt.clf()
                chow_obj.plot(
                    filename=f'{"/".join([savingpath, chow_obj.name])}.{fmt}',
                    title=params.get('prefix', '') + chow_obj.name,
                    **params)
                pbar.update(1)
                plt.close()
    def plot_by_group(self,
                      savingpath:Union[None,str] = 'figs',
                      fmt:str = 'pdf',
                      cmap:str = 'Set1',
                      **params):
        """
        Plot all trends in individual files, grouped by category.
        
        :param savingpath: Path where the plots will be saved
                           (None to show instead of save)
        :param fmt: Format of the image (e.g., 'pdf', 'png')
        :param cmap: Name of one of matplotlib cmaps
        :param params: Named parameters passed to matplotlib and plot_by_group.
                      Includes: show_legend, plot_overall, plot_individual_fill,
                      xlabel, ylabel, title, overall_label
        """
        if not os.path.isdir(savingpath):
            os.makedirs(savingpath)
        print("Plotting trends by group")
        with tqdm(total = len(self.chows)) as pbar:
            for chow_obj in self.chows:
                plt.clf()
                params['title'] = params.get('title', chow_obj.name)
                params['xlabel'] = params.get('xlabel', self.timecol)
                params['ylabel'] = params.get('ylabel', self.ycol)
                if savingpath is not None:
                    filename = f"{savingpath}/{chow_obj.name}_bygroups.{fmt}"
                else:
                    filename = None
                chow_obj.plot_by_group(
                    filename=filename, groupcol=self.groupcol, cmap=cmap, **params
                )
                pbar.update(1)
        if savingpath is None:
            plt.show()
            plt.close()

    def results_df(self):
        """
        Create dataframe with analysis results.
        
        :return: DataFrame with multi-level columns containing results
                 for all analyzed series
        """
        d = {'name': [],
             'breakpoint': [],
             'score': [],
             'model0': [],
             'model1': [],
             'model2': []
             }
        model_params = self.chows[0].params_names()
        for c in self.chows:
            params = c.params()
            d['name'].append(c.name)
            d['breakpoint'].append(c.initial_breakpoint)
            d['score'].append(c.score)
            for k,v in params.items():
                d[k].append(v)
        data, cols = [],[]
        for k, v in d.items():
            if not isinstance(v[0], dict):
                cols.append((k, ''))
                data.append(v)
            else:
                subdata = {}
                for v1 in v:
                    for p in model_params:
                        subdata.setdefault(p,[]).append(v1.get(p,np.nan))
                for k1,v1 in subdata.items():
                    cols.append((k, k1))
                    data.append(v1)
        df = pd.DataFrame(list(zip(*data)), columns=pd.MultiIndex.from_tuples(cols))
        return df
    def to_excel(self, filename:str, **kwargs):
        """
        Export results to Excel file.
        
        Creates multiple sheets: 'results' (analysis results), 
        'data' (input data),
        and 'failed' (list of failed analyses).

        :param filename: Output filename with .xlsx or .xls extension
        :param kwargs: Additional keyword arguments passed to pd.ExcelWriter
        """
        df = self.results_df()
        with pd.ExcelWriter(filename, **kwargs) as writer:
            df.to_excel(writer, sheet_name = 'results')
            self.df.to_excel(writer, sheet_name = 'data')
            pd.DataFrame([[x] for x in self.failed],columns=['Name']).to_excel(writer, sheet_name='failed')
    def to_csv(self, filename:str, **kwargs):
        """
        Export results to CSV file.

        :param filename: Output filename with .csv extension
        :param kwargs: Additional keyword arguments passed to pd.DataFrame.to_csv
        """

        df = self.results_df()
        df.to_csv(filename, **kwargs)
    def save_to_file(self, filename:str, **kwargs):
        """
        Export results to Excel or CSV file based on extension.
        
        Automatically detects file format from extension and 
        calls the appropriate export method.

        :param filename: Output filename with extension (.xlsx, .xls, or .csv)
        :param kwargs: Additional keyword arguments passed to export methods
        :raises ValueError: If filename extension is not supported
        """

        ext = filename.split('.')[-1]
        if ext in ['xlsx','xls']:
            self.to_excel(filename, engine='openpyxl')
        elif ext == 'csv':
            self.to_csv(filename,**kwargs)
        else:
            raise ValueError(f'filename {filename} does not have a supported extension. Should be .csv, .xlsx or .xls.')



"""Utility functions and constants for the chowclassifier package."""

from typing import Union
from datetime import datetime
import pandas as pd
from matplotlib import colormaps

TREND_COLORS = ["#4d4c49", "#d85e5e", "#3a668c", "#000000"]  # constant, increasing, decreasing
CI_COLORS = ["#fef5da", "#ff988c", "#b4c5f4", "#474747"]
SCATTERPARAMS = {"marker": 'o', "linewidths":0.5, "alpha":1,"s":50,"label":'_nolegend_'}
FILLPARAMS = {"linestyle":"None", "alpha":0.3,"label":'_nolegend_'}
FITPARAMS = {"label":'_nolegend_'}

def parse_cmap(cmap):
    """
    Parse and normalize colormap to dictionary format.
    
    Converts string colormap names, lists, or dictionaries to a standardized
    dictionary format with integer keys.
    
    :param cmap: Colormap as string name, list of colors, or dictionary
    :return: Dictionary mapping integers to color values
    """
    if isinstance(cmap,str):
        cmap = dict(enumerate(colormaps[cmap].colors))
    elif isinstance(cmap,list):
        cmap = dict(enumerate(cmap))
    elif isinstance(cmap, dict):
        if not all(isinstance(item, int) for item in cmap.keys()):
            cmap.update({i:v for i,(k,v) in enumerate(cmap.items())})
    else:
        cmap = dict(enumerate(colormaps['Set1'].colors))
    return cmap

def conf_int_to_trend(ci:Union[pd.DataFrame,list], param:str='X'):
    """
    Convert confidence interval to trend classification.
    
    Returns trend code 0, 1, or 2 from the confidence interval:

    - if ci crosses 0, then returns 0 (non-significant)
    - if ci is completely below 0, then returns 2 (decreasing)
    - else returns 1 (increasing)
    
    :param ci: Confidence interval to process, either a tuple (length 2), 
               a DataFrame, or result from conf_int given by statsmodels
    :param param: Column name to use if a DataFrame was given as ci
    :return: Integer trend code (0, 1, or 2)
    """
    if isinstance(ci, list) and len(ci)==2:
        l,u = ci
    else:
        l,u=ci.loc[param,:]
    if u<0:
        return 2 # significantly decreasing (at tested level)
    if l>0:
        return 1 # significantly increasing (at tested level)
    return 0 # non significant (at tested level)

def infer_date_format(s_date:str):
    """
    Infer date format from a sample date string.

    :param s_date: Sample date string to infer format from
    :return: Date format string if successfully inferred, None otherwise
    """
    date_patterns = ["%d-%m-%Y","%d-%m-%Y %H:%M:%S",
                     "%Y-%m-%d","%Y-%m-%d %H:%M:%S",
                     "%y-%m-%d","%y-%m-%d %H:%M:%S",
                     "%d %b %y","%d %b %y %H:%M:%S",
                     "%d %b %Y","%d %b %Y %H:%M:%S",
                     ]

    for pattern in date_patterns:
        try:
            datetime.strptime(s_date, pattern).date()
        except ValueError:
            pass
        else:
            return pattern
    return None

def guess_timelapse_type(series:Union[list,pd.DataFrame,pd.Series], date_format:str = "%Y-%m-%d"):
    """
    Guess the time lapse unit in a time series and re-index it.
    
    Analyzes datetime differences to determine whether data is indexed by
    seconds, minutes, hours, days, or years.

    :param series: Time series data (list, DataFrame, or Series)
    :param date_format: Format string for parsing datetime values
    :return: Tuple of (indexed series, datetype string)
    """
    try:
        series = pd.to_datetime(series, format = date_format)
    except (ValueError, TypeError):
        pass
    start = min(series)
    series = series-start # reindex timecol from lowest one
    if all(i.seconds==0 for i in series):
        #check if indexed by minutes
        if all(i.seconds%60 == 0 for i in series):
            #check if indexed by hour
            if all(i.seconds%3600 == 0 for i in series):
                #check if indexed by year
                if all(i.days%365==0 or i.days%366==0 for i in series):
                    #index by year
                    series = [int(i.days/365) for i in series]
                    datetype = 'year'
                else:#index by day
                    series = [i.days for i in series]
                    datetype = 'day'
            else:#index by hour
                series = [i.days * 24 + i.hours for i in series]
                datetype = 'hour'
        else:#index by minute
            series = [i.days * 60 * 24 + i.minutes for i in series]
            datetype = 'minute'
    else:# index on the seconds:
        series = [i.days * 3600 * 24 + i.seconds for i in series]
        datetype = 'second'
    return series,datetype
